from gustaf.utils import arr
from gustaf.utils import connec
from gustaf.utils import log
from gustaf.utils import tictoc

# Alias
connectivity = connec

__all__ = [
        "arr",
        "connec",
        "connectivity",
        "log",
        "tictoc",
]
